#include <linux/i2c.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include "innopmbus_drv.h"
#include "hal_interface.h"

#include <linux/cpufreq.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/version.h>
#include <linux/thermal.h>

#include "inno_timer.h"
#include "inno_io.h"
#include "hal.h"
#include "innopower.h"
#include "inno_devfreq_gov.h"
#include "inno_input_event.h"

#define INNO_POWERCHIP_MAX_NUM	32

static struct i2c_device_id inno_powerchip_id_tbl[INNO_POWERCHIP_MAX_NUM];
static int id_tbl_idx = 0;
static struct i2c_driver *i2c_drv = NULL;

static uint32_t innopwr_get_vol_cmd_ver(struct dev_rsrc *pdevrsrc)
{
	return pdevrsrc->vol_cmd_version;
}

#define VOL_CMD_HOLD_10MS (10000)
static int innopwr_drv_mcufw_sysstat_notify(struct innopower *power, uint32_t sys_stat, uint32_t timout)
{
	int err = 0;
	struct dev_rsrc *pdev_rsrc = inno_rsrc_devres_find(power->pdev);
	uint32_t vol_cmd_verison = innopwr_get_vol_cmd_ver(pdev_rsrc);

	if (pdev_rsrc && pdev_rsrc->mcufw_comm_v) {
		if (vol_cmd_verison >= MCUFW_COMM_CMD_VOL_V2) {
			err = hal_mcufw_comm_msg_xfer(pdev_rsrc, MCUFW_COMM_MODULE_VOLTAGE, MCUFW_COMM_CMD_VOL_NOTIFY, &sys_stat, sizeof(sys_stat));
			if (err) {
				innopwr_notice("run finished err[%d] verison[%d %u] sys_stat[%d]\n",
					err, pdev_rsrc->mcufw_comm_v, vol_cmd_verison, sys_stat);
			}
			/*The interval time between commands to avoid mcufw lost irqs*/
			if (timout) {
				inno_usleep_range(timout, timout);
			}
		}
	}

	return err;
}

static int innopwr_drv_mcufw_set_voltage(void* chip_ctx, uint32_t mv, int sys_stat)
{
	uint32_t cmd_data[5] = {0};
	uint32_t vol_cmd_verison = innopwr_get_vol_cmd_ver(chip_ctx);

	if (MCUFW_COMM_CMD_VOL_V1 == vol_cmd_verison) {
		innopwr_dbg("MCUFW_COMM_CMD_VOL_V1 mv[%u] with sys_stat[%d]\n", mv, sys_stat);
		cmd_data[0] = mv;
		cmd_data[1] = sys_stat;
		return hal_mcufw_comm_msg_xfer(chip_ctx, MCUFW_COMM_MODULE_VOLTAGE, MCUFW_COMM_CMD_VOL_SET, cmd_data, sizeof(cmd_data));
	} else {
		innopwr_dbg("MCUFW_COMM_CMD_VOL_V%u mv[%u] without sys_stat[%d]\n", vol_cmd_verison, mv, sys_stat);
		return hal_mcufw_comm_msg_xfer(chip_ctx, MCUFW_COMM_MODULE_VOLTAGE, MCUFW_COMM_CMD_VOL_SET, &mv, sizeof(mv));
	}
}

static int innopwr_drv_mcufw_get_voltage_from_bmc(void* chip_ctx, uint32_t *mv)
{
	return hal_get_voltage_from_bmc(chip_ctx, mv);
}

static int innopwr_drv_mcufw_get_voltage_from_reg(void* chip_ctx, uint32_t *mv)
{
	return hal_mcufw_comm_msg_xfer(chip_ctx, MCUFW_COMM_MODULE_VOLTAGE, MCUFW_COMM_CMD_VOL_GET, mv, sizeof(*mv));
}

static int innopwr_drv_mcufw_get_voltage(void* chip_ctx, uint32_t *mv)
{
	if (PWRD_DBG_REAL_VOL == get_pwr_debug_lvl()) {
		return innopwr_drv_mcufw_get_voltage_from_reg(chip_ctx, mv);
	} else {
		return innopwr_drv_mcufw_get_voltage_from_bmc(chip_ctx, mv);
	}
}

static int innopower_set_voltage(void *chip_ctx, unsigned int vol, int sys_stat)
{
	int err = 0;
	struct dev_rsrc *pdev_rsrc = (struct dev_rsrc *)chip_ctx;
	struct innopower *power = platform_get_drvdata(pdev_rsrc->power_dev[0]);

	if (!is_support_update_voltage() || !hal_vol_is_digital(pdev_rsrc->dev)) {
		innopwr_dbg("is_support_update_vol[%d] is_digital_vol[%d]\n",
			is_support_update_voltage(), hal_vol_is_digital(pdev_rsrc->dev));
		return INNO_POWER_OK;
	}

	if (!innopwr_get_vol_cmd_ver(pdev_rsrc)) {
		if (!power || !power->chip || !power->chip->set_gpu_voltage) {
			innopwr_notice("null pointer\n");
			return INNO_POWER_ERR;
		}

		return power->chip->set_gpu_voltage(power->chip, 0, vol);
	} else {
		err = innopwr_drv_mcufw_set_voltage(chip_ctx, vol, sys_stat);
		if (err) {
			innopwr_notice("set voltage is failed err[%d]\n", err);
			return INNO_POWER_ERR;
		}
	}

	return INNO_POWER_OK;
}

static int innopower_get_voltage(void *chip_ctx)
{
	struct dev_rsrc *pdev_rsrc = (struct dev_rsrc *)chip_ctx;
	struct innopower *power = platform_get_drvdata(pdev_rsrc->power_dev[0]);
	int vol = 0;

	if (!hal_vol_is_digital(pdev_rsrc->dev)) {
		return ANALOG_VOLT_DEFAULT_OUTPUT;
	}

	if (!innopwr_get_vol_cmd_ver(pdev_rsrc)) {
		if (!power || !power->chip || !power->chip->get_gpu_voltage) {
			innopwr_notice("null pointer\n");
			return INNO_POWER_ERR;
		}

		vol = power->chip->get_gpu_voltage(power->chip, 0);
		return vol;
	} else {
		if (!innopwr_drv_mcufw_get_voltage(chip_ctx, (uint32_t *)&vol)) {
			innopwr_info("get mv = %d\n", vol);
			return vol;
		} else {
			innopwr_notice("get failed\n");
			return INNO_POWER_ERR;
		}
	}
}

ssize_t innopower_store_gpu_voltage(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	int retval = 0;
	int gpu_voltage = 0;
	struct innopower *power = NULL;
	struct dev_rsrc *pdev_rsrc = NULL;
	
	power = platform_get_drvdata(to_platform_device(dev));
	if (!power){
		innopwr_notice("null pointer\n");
		return -EINVAL;
	}

	pdev_rsrc = inno_rsrc_devres_find(power->pdev);
	if (!pdev_rsrc){
		innopwr_notice("null pointer\n");
		return -EINVAL;
	}

	if (!is_support_update_voltage()) {
		return count;
	}

	if (sscanf(buf, " %i ", &gpu_voltage) != 1) {
		innopwr_err("Failed to store  gpu_voltage attribute\n");
		return -EINVAL;
	}

	if (!innopwr_get_vol_cmd_ver(pdev_rsrc)) {
		if (!power || !power->chip || !power->chip->set_gpu_voltage) {
			innopwr_notice("null pointer\n");
			return -EINVAL;
		}

		retval = power->chip->set_gpu_voltage(power->chip, 0, gpu_voltage);
	} else {
			hal_set_voltage(power->pdev, (unsigned int)gpu_voltage);
	}
	innopwr_notice("gpu_voltage = %d\n", gpu_voltage);

	return retval ? retval : count;
}

ssize_t innopower_show_gpu_voltage(struct device *dev, struct device_attribute *attr, char *buf)
{
	int  retval = 0;
	int  gpu_voltage = 0;
	int  retry = 0;
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));
	struct dev_rsrc *pdev_rsrc = inno_rsrc_devres_find(power->pdev);

	do {
		retval = innopower_get_voltage(pdev_rsrc);
		retry++;
		inno_msleep(5);
	} while ((retry < 5) && (retval <= 0 || retval > 1200));

	if (retval <= 0 || retval > 1200) {
		innopwr_err("Failed to read  gpu_voltage attribute\n");
		sprintf(buf, "0\n");
		return retval;
	}
	gpu_voltage = retval;
	innopwr_notice("[%s]: gpu_voltage = %d retry=[%d]\n",
		gpu_voltage > 950 ? "warnning" : "success", retval, retry);

	return sprintf(buf, "%d\n", gpu_voltage);
}

DEVICE_ATTR(gpu_voltage, S_IRUGO | S_IWUSR, innopower_show_gpu_voltage, innopower_store_gpu_voltage);

ssize_t innopower_store_pwrhelp(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	return count;
}

ssize_t innopwr_show_pwrhelp(struct device *dev, struct device_attribute *attr, char *buf)
{
	ssize_t count = 0;
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));
	struct dev_rsrc *pdev_rsrc = inno_rsrc_devres_find(power->pdev);
	mcufw_ver_info_t *pmcufw_ver = &pdev_rsrc->mcufw_ver_info;
	mcufw_ver_desc_t * pmcufw_ver_desc = &pmcufw_ver->mcufw_ver_desc;
	uint32_t vol_cmd_version = pdev_rsrc->vol_cmd_version;

	count += inno_sprintf(buf + count, 128, "mcufw %ubit format, %s verison[%u.%u.%u.%u]\n",
		pmcufw_ver->mcufw_v_format, pmcufw_ver->internal_verison ? "internal" : "external" , pmcufw_ver_desc->mcufw_major,
		pmcufw_ver_desc->mcufw_minor, pmcufw_ver_desc->mcufw_revision, pmcufw_ver_desc->mcufw_build);

	count += inno_sprintf(buf + count, 128, "kmd2mcufw mcufw_comm_v[%d] vol_cmd_version[%u]: bmcv[%d]",
		pdev_rsrc->mcufw_comm_v, pdev_rsrc->vol_cmd_version, pdev_rsrc->bmc.verison);

	if (MCUFW_COMM_CMD_VOL_V1 == vol_cmd_version) {
		count += inno_sprintf(buf + count, 128, "updating voltage with sys_stat\n");
	} else if (vol_cmd_version >= MCUFW_COMM_CMD_VOL_V2) {
		count += inno_sprintf(buf + count, 128, "updating voltage with not sys_stat and additional cmd is used to do it\n");
	} else {
		count += inno_sprintf(buf + count, 128, "kmd Fallback to legacy: host drv access voltage chip by pmbus directly\n");
	}

	return count;
}

DEVICE_ATTR(pwrhelp, S_IRUGO | S_IWUSR, innopwr_show_pwrhelp, innopower_store_pwrhelp);

static int innopwr_pmbus_rs_lock(struct powerchip *pwrchip, bool chip_need_rs)
{
	struct inno_pmbus_dev *pmbus = NULL;
	pmbus = (struct inno_pmbus_dev *)pwrchip->priv_data;
	if (!pmbus) {
		innopwr_err("pmbus get failed!!!\n");
		return INNO_POWER_ERR;
	}
	mutex_lock(&pmbus->chip_rs_lock);
	pmbus->chip_need_rs = chip_need_rs;
	return INNO_POWER_OK;
}

static int innopwr_pmbus_rs_unlock(struct powerchip *pwrchip)
{
	struct inno_pmbus_dev *pmbus = NULL;
	pmbus = (struct inno_pmbus_dev *)pwrchip->priv_data;
	if (!pmbus) {
		innopwr_err("pmbus get failed!!!\n");
		return INNO_POWER_ERR;
	}
	pmbus->chip_need_rs = 1;
	mutex_unlock(&pmbus->chip_rs_lock);
	return INNO_POWER_OK;
}

static int _innopwr_read8(struct powerchip *pwrchip, unsigned char reg, bool chip_need_rs)
{
	int ret = 0;
	if (!pwrchip) {
		innopwr_notice("gchip is null!!!\n");
		return INNO_POWER_ERR;
	}

	mutex_lock(&pwrchip->lock);
	if (innopwr_pmbus_rs_lock(pwrchip, chip_need_rs)) {
		mutex_unlock(&pwrchip->lock);
		innopwr_notice("%s read reg[0x%x] pmbus_rs_lock failed!!!\n", pwrchip->name, reg);
		return INNO_POWER_ERR;
	}

	ret = i2c_smbus_read_byte_data(pwrchip->client, reg);
	if (ret < 0) {
		innopwr_pmbus_rs_unlock(pwrchip);
		mutex_unlock(&pwrchip->lock);
		innopwr_err("%s read reg[0x%x] failed!!!\n", pwrchip->name, reg);

		return INNO_POWER_ERR;
	}
	innopwr_pmbus_rs_unlock(pwrchip);
	mutex_unlock(&pwrchip->lock);

	return ret;
}

static int _innopwr_write8(struct powerchip *pwrchip, unsigned char reg, unsigned char data, bool chip_need_rs)
{
	int ret = 0;

	if (!pwrchip) {
		innopwr_notice("gchip is null!!!\n");
		return INNO_POWER_ERR;
	}

	mutex_lock(&pwrchip->lock);
	if (innopwr_pmbus_rs_lock(pwrchip, chip_need_rs)) {
		mutex_unlock(&pwrchip->lock);
		innopwr_notice("%s write 0x%x to reg[0x%x] pmbus_rs_lock failed!!!\n", pwrchip->name, data, reg);
		return INNO_POWER_ERR;
	}
	ret = i2c_smbus_write_byte_data(pwrchip->client, reg, data);
	if (ret < 0) {
		innopwr_pmbus_rs_unlock(pwrchip);
		mutex_unlock(&pwrchip->lock);
		innopwr_err("%s write 0x%x to reg[0x%x] failed!!!\n", pwrchip->name, data, reg);

		return INNO_POWER_ERR;
	}
	innopwr_pmbus_rs_unlock(pwrchip);
	mutex_unlock(&pwrchip->lock);

	return INNO_POWER_OK;
}

static int _innopwr_read16(struct powerchip *pwrchip, unsigned char reg, bool chip_need_rs)
{
	int ret = 0;

	if (!pwrchip) {
		innopwr_notice("gchip is null!!!\n");
		return INNO_POWER_ERR;
	}

	mutex_lock(&pwrchip->lock);
	if (innopwr_pmbus_rs_lock(pwrchip, chip_need_rs)) {
		mutex_unlock(&pwrchip->lock);
		innopwr_notice("%s read reg[0x%x] pmbus_rs_lock failed!!!\n", pwrchip->name, reg);
		return INNO_POWER_ERR;
	}
	ret = i2c_smbus_read_word_data(pwrchip->client, reg);
	if (ret < 0) {
		innopwr_pmbus_rs_unlock(pwrchip);
		mutex_unlock(&pwrchip->lock);
		innopwr_err("%s read reg[0x%x] failed!!!\n", pwrchip->name, reg);

		return INNO_POWER_ERR;
	}
	innopwr_pmbus_rs_unlock(pwrchip);
	mutex_unlock(&pwrchip->lock);

	return ret;
}

static int _innopwr_write16(struct powerchip *pwrchip, unsigned char reg, unsigned int data, bool chip_need_rs)
{
	int ret = 0;

	if (!pwrchip) {
		innopwr_notice("gchip is null!!!\n");
		return INNO_POWER_ERR;
	}

	mutex_lock(&pwrchip->lock);
	if (innopwr_pmbus_rs_lock(pwrchip, chip_need_rs)) {
		mutex_unlock(&pwrchip->lock);
		innopwr_notice("%s write 0x%x to reg[0x%x] pmbus_rs_lock failed!!!\n", pwrchip->name, data, reg);
		return INNO_POWER_ERR;
	}
	ret = i2c_smbus_write_word_data(pwrchip->client, reg, data);
	if (ret < 0) {
		innopwr_pmbus_rs_unlock(pwrchip);
		mutex_unlock(&pwrchip->lock);
		innopwr_err("%s write 0x%x to reg[0x%x] failed!!!\n", pwrchip->name, data, reg);

		return INNO_POWER_ERR;
	}
	innopwr_pmbus_rs_unlock(pwrchip);
	mutex_unlock(&pwrchip->lock);

	return ret;
}

int innopwr_read8(struct powerchip *pwrchip, unsigned char reg)
{
	return _innopwr_read8(pwrchip, reg, 1);
}

int innopwr_read8_no_repeat_start(struct powerchip *pwrchip, unsigned char reg)
{
	return _innopwr_read8(pwrchip, reg, 0);
}

int innopwr_write8(struct powerchip *pwrchip, unsigned char reg, unsigned char data)
{
	return _innopwr_write8(pwrchip, reg, data, 1);
}

int innopwr_write8_no_repeat_start(struct powerchip *pwrchip, unsigned char reg, unsigned char data)
{
	return _innopwr_write8(pwrchip, reg, data, 0);
}

int innopwr_read16(struct powerchip *pwrchip, unsigned char reg)
{
	return _innopwr_read16(pwrchip, reg, 1);
}

int innopwr_read16_no_repeat_start(struct powerchip *pwrchip, unsigned char reg)
{
	return _innopwr_read16(pwrchip, reg, 0);
}

int innopwr_write16(struct powerchip *pwrchip, unsigned char reg, unsigned int data)
{
	return _innopwr_write16(pwrchip, reg, data, 1);
}

int innopwr_write16_no_repeat_start(struct powerchip *pwrchip, unsigned char reg, unsigned int data)
{
	return _innopwr_write16(pwrchip, reg, data, 0);
}

static int innopower_sys_register(struct device *dev)
{
	int ret = 0;

	ret = device_create_file(dev, &dev_attr_gpu_voltage);

	ret = device_create_file(dev, &dev_attr_pwrhelp);

	return ret;
}

static void innopower_sys_unregister(struct device *dev)
{
	device_remove_file(dev, &dev_attr_pwrhelp);

	device_remove_file(dev, &dev_attr_gpu_voltage);
}

#define I2C_SLAVE_ADDR_VALID 0x7F
#define I2C_SLAVE_ADDR_INVALID 0x0
#define I2C_SLAVE_ADDE_DEFAULT_VALID 0x20
/*default double chip use xdpe12284c*/
#define DOUBLE_CHIP_IDX0_DEFAULT_ADDR 0xEC
static int innopower_register_chip(struct innopower *power)
{
	struct powerchip *chip;
	struct i2c_board_info info;

	chip = devm_kzalloc(power->pltdev, sizeof(struct powerchip), GFP_KERNEL);
	if (!chip) {
		innopwr_err("devm_kzalloc failed\n");
		goto chip_fail;
	}

	/*i2c info static register*/
	/*double is not support hwinfo and custom info parase, and have a default addr*/

	memset(&info, 0, sizeof(struct i2c_board_info));
	if (power->params->slave_addr == I2C_SLAVE_ADDR_INVALID) {
		power->params->slave_addr = DOUBLE_CHIP_IDX0_DEFAULT_ADDR ;
		if (power->params->slave_addr > I2C_SLAVE_ADDR_VALID) {
			info.addr = power->params->slave_addr >> 1;
		} else {
			info.addr = power->params->slave_addr;
		}

		strcpy(power->params->chip_name, "VIRTURE_CHIP");
	}

	info.addr = power->params->slave_addr;
	if (info.addr > I2C_SLAVE_ADDR_VALID) { /*default xdpe12284c*/
		info.addr = DOUBLE_CHIP_IDX0_DEFAULT_ADDR >> 1;
	}
	sprintf(info.type, "%s-%d", power->params->chip_name, id_tbl_idx);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 2, 0))
	chip->client = i2c_new_device(power->adapter, &info);
#else
	chip->client = i2c_new_client_device(power->adapter, &info);
#endif
	if (!chip->client) {
		innopwr_err("i2c_new_device failed\n");
		goto chip_fail;
	}
	i2c_set_clientdata(chip->client, chip);
	power->chip = chip;
	chip->priv_data = i2c_get_adapdata(power->adapter);

	/*inno_powerchip_id_tbl is created dynamic*/
	strcpy(inno_powerchip_id_tbl[id_tbl_idx].name, chip->client->name);
	id_tbl_idx++;

	i2c_drv = innopower_get_drvobj(power->params->chip_name);
	if (!i2c_drv) {
		innopwr_err("not support drv\n");
		goto chip_fail;
	}
	innopwr_info("name = %s type= %s addr = 0x%x i2c_drv = %px\n", chip->client->name, info.type, info.addr, i2c_drv);

	return 0;

chip_fail:
	devm_kfree(power->pltdev, chip);
	power->chip = NULL;

	return -ENOMEM;
}

static int innopower_unregister_chip(struct innopower *power)
{
	if (power->chip) {
		i2c_unregister_device(power->chip->client);
		devm_kfree(power->pltdev, power->chip);
	}

	return 0;
}

static uint32_t innopwr_find_vol(inno_dev *dev, uint32_t gpu_freq)
{
	int i = 0;
	uint32_t vol = 0;
	struct inno_gpu_opp *opp_tbl = hal_power_get_opptbl(dev);
	int32_t opp_size = hal_power_get_opptbl_size(dev);

	for (i = 0; i < opp_size; i++) {
		if (gpu_freq == opp_tbl[i].freq / GPU_FREQ_M2HZ_UNIT) {
			vol = opp_tbl[i].volt / GPU_VOL_UV2MV_UNIT;
			break;
		}
	}

	if (!vol) {
		vol = opp_tbl[opp_size-1].volt / GPU_VOL_UV2MV_UNIT;
		innopwr_notice("found init vol failed and take max vol[%u] as init vol opp_size[%d]\n", vol, opp_size);
	}

	return vol;
}

static int innopower_vol_init(inno_dev *dev, int sys_stat)
{
	struct dev_rsrc *pdev_rsrc = inno_rsrc_devres_find(dev);
	struct power_custom_hwinfo_t *pwrinfo = &pdev_rsrc->pwrinfo;
	uint8_t  pwr_default_mode = pwrinfo->pwr_ctrl_mode & INNO_POWER_DEFAULT_WORKMODE_MASK;
	uint32_t init_vol = 0;
	uint16_t init_freq = 0;

	if (!is_support_update_voltage()) {
		return 0;
	}

	switch (pwr_default_mode) {
		case DEFAULT_WORKMODE_POWSAVE:
			init_freq = pwrinfo->powersave_mode_freq;
			break;
		case DEFAULT_WORKMODE_DYN:
			init_freq = pwrinfo->dyn_mode_mode_freq;
			break;
		case DEFAULT_WORKMODE_PERF:
			init_freq = pwrinfo->performance_mode_freq;
			break;
		case DEFAULT_WORKMODE_NORMAL:
		default:
			init_freq = pwrinfo->normal_mode_freq;
			break;
	}

	init_vol = innopwr_find_vol(dev, init_freq);
	innopwr_dbg("pwr_default_mode = %u init_freq[%d] init_vol[%u]\n", pwr_default_mode, init_freq, init_vol);

	if (INNO_POWER_OK != innopower_set_voltage(pdev_rsrc, init_vol, sys_stat)) {
		innopwr_err("set hw init voltage failed and pwr_default_mode=[%u] init_freq=[%d] init_vol[%u]\n",
			pwr_default_mode, init_freq, init_vol);
		return INNO_POWER_ERR;
	}

	return INNO_POWER_OK;
}

static int innopower_pmbus_probe(struct platform_device *pdev)
{
	struct innopower *power;
	struct device *dev = &pdev->dev;
	struct dev_rsrc *pdev_rsrc = NULL;
	int init_vol = 0;

	power = devm_kzalloc(dev, sizeof(*power), GFP_KERNEL);
	if (!power) {
		innopwr_err("devm_kzalloc failed\n");
		goto power_fail;
	}

	power->pltdev = dev;
	platform_set_drvdata(pdev, power);
	power->params = hal_get_volctrl_hwinfo(pdev->dev.parent);
	power->adapter = hal_get_pmbus_adapter(pdev->dev.parent, power->params->pmbus_id);
	if (!power->adapter) {
		innopwr_err("i2c_get_adapter failed\n");
		goto power_fail;
	}

	innopower_register_chip(power);

	power->pdev = pdev->dev.parent;
	pdev_rsrc = inno_rsrc_devres_find(power->pdev);
	init_vol = innopower_get_voltage(pdev_rsrc);
	hal_set_voltage_ops(power->pdev, innopower_set_voltage, innopower_get_voltage);
	innopower_vol_init(power->pdev, PWR_NOTIFY_FOR_NORMAL);
	innopower_sys_register(&pdev->dev);

	innopwr_dbg("run finished and init vol is %d is_digital = %d is_support_update_voltage = %d\n",
		init_vol, hal_vol_is_digital(power->pdev), is_support_update_voltage());

	return 0;

power_fail:
	devm_kfree(dev, power);
	innopwr_err("power_fail\n");

	return -ENOMEM;
}

static int innopower_pmbus_remove(struct platform_device *pdev)
{
	struct innopower *power = platform_get_drvdata(pdev);
	struct device *dev = &pdev->dev;

	innopower_sys_unregister(&pdev->dev);

	innopower_vol_init(power->pdev, PWR_NOTIFY_FOR_NORMAL);

	innopower_unregister_chip(power);

	devm_kfree(dev, power);

	innopwr_dbg("run finished\n");

	return 0;
}

static void innopower_pmbus_shutdown(struct platform_device *pdev)
{
	struct innopower *power = platform_get_drvdata(pdev);

	/* fix bug11626/11628: "magic num or version num error" is
	 * reported by mcufw as mcufw cleaned protocal vram */
	innopower_vol_init(power->pdev, PWR_NOTIFY_FOR_SHUTDOWN);

	inno_usleep_range(10000, 10000);

	innopwr_drv_mcufw_sysstat_notify(power, PWR_NOTIFY_FOR_SHUTDOWN, 0);

	innopwr_dbg("run finished\n");
}
struct platform_device_id innopower_pmbus_device_id_table[] = {
	{ .name = INNO_POWER_DEVICE_NAME, .driver_data = 0 },
	{},
};

static int innopwr_enter_sleep(struct device *dev, char *msg)
{
	int ret = 0;
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));
	struct dev_rsrc *pdev_rsrc = inno_rsrc_devres_find(power->pdev);
	uint32_t sleep_gpu_freq = hal_get_pll(pdev_rsrc->dev, PLL_GPU);
	uint32_t real_vol = 0;

	if (!is_support_update_voltage()) {
		return 0;
	}

	if (hal_vol_is_digital(pdev_rsrc->dev)) {
		if (innopwr_drv_mcufw_get_voltage_from_reg(pdev_rsrc, &real_vol)) {
			innopwr_notice("%s read voltage failed and sleep_vol[%d]\n", msg, pdev_rsrc->sleep_vol);
			return -1;
		}

	} else {
		real_vol = ANALOG_VOLT_DEFAULT_OUTPUT;
	}
	pdev_rsrc->sleep_vol = innopwr_find_vol(pdev_rsrc->dev, sleep_gpu_freq);

	pdev_rsrc->sleep_freq = sleep_gpu_freq;
	innopwr_dbg("%s run finished and opp[%u, %d %u]\n", msg, sleep_gpu_freq, pdev_rsrc->sleep_vol, real_vol);

	return ret;
}

static int innopwr_exit_sleep(struct device *dev, char *msg, int sys_stat)
{
	int ret = 0;
	uint32_t real_vol = 0;
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));
	struct dev_rsrc *pdev_rsrc = inno_rsrc_devres_find(power->pdev);
	uint32_t gpu_freq = hal_get_pll(pdev_rsrc->dev, PLL_GPU);

	if (!is_support_update_voltage()) {
		return 0;
	}

	if (get_pwr_debug_lvl() >= PWRD_LVL_DBG) {
		if (hal_vol_is_digital(pdev_rsrc->dev)) {
			if (!innopwr_drv_mcufw_get_voltage_from_reg(pdev_rsrc, &real_vol)) {
				innopwr_notice("%s bf run start and opp[%u, %u | %u, %d]\n",
						msg, gpu_freq, real_vol, pdev_rsrc->sleep_freq, pdev_rsrc->sleep_vol);
			} else {
				innopwr_notice("%s bf read voltage failed\n", msg);
			}
		} else {
			real_vol = ANALOG_VOLT_DEFAULT_OUTPUT;
		}
	}

	if (pdev_rsrc->sleep_vol > 0) {
		if (gpu_freq < pdev_rsrc->sleep_freq) { //increase freq
			innopower_set_voltage(pdev_rsrc, pdev_rsrc->sleep_vol, sys_stat);
		}

		hal_set_pll(pdev_rsrc->dev, PLL_GPU, pdev_rsrc->sleep_freq);

		if (gpu_freq >= pdev_rsrc->sleep_freq) { //decrease freq
			innopower_set_voltage(pdev_rsrc, pdev_rsrc->sleep_vol, sys_stat);
		}
	}

	if (get_pwr_debug_lvl() >= PWRD_LVL_DBG) {
		gpu_freq = hal_get_pll(pdev_rsrc->dev, PLL_GPU);
		if (hal_vol_is_digital(pdev_rsrc->dev)) {
			if (!innopwr_drv_mcufw_get_voltage_from_reg(pdev_rsrc, &real_vol)) {
				innopwr_notice("%s af run finished and opp[%u, %u | %u, %d]\n",
						msg, gpu_freq, real_vol, pdev_rsrc->sleep_freq, pdev_rsrc->sleep_vol);
			} else {
				innopwr_notice("%s af read voltage failed\n", msg);
			}
		} else {
			innopwr_notice("%s bf analog power real_vol is %d\n", msg, real_vol);
		}
	}

	return ret;
}

static int innopwr_suspend(struct device *dev)
{
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));

	innopwr_drv_mcufw_sysstat_notify(power, PWR_NOTIFY_FOR_S3_ENTER, VOL_CMD_HOLD_10MS);

	innopwr_enter_sleep(dev, "s3 suspend");

	return 0;
}

static int innopwr_resume(struct device *dev)
{
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));

	innopwr_drv_mcufw_sysstat_notify(power, PWR_NOTIFY_FOR_S3_EXIT, VOL_CMD_HOLD_10MS);

	innopwr_exit_sleep(dev, "s3 resume", PWR_NOTIFY_FOR_S3_EXIT);

	return 0;
}

static int innopwr_freeze(struct device *dev)
{
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));

	innopwr_drv_mcufw_sysstat_notify(power, PWR_NOTIFY_FOR_S4_ENTER, VOL_CMD_HOLD_10MS);

	innopwr_enter_sleep(dev, "s4 freeze");

	return 0;
}

static int innopwr_thaw(struct device *dev)
{
	int ret = 0;

	innopwr_dbg("run finished\n");

	return ret;
}
static int innopwr_poweroff(struct device *dev)
{
	int ret = 0;
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));

	innopwr_drv_mcufw_sysstat_notify(power, PWR_NOTIFY_FOR_POWEROFF, VOL_CMD_HOLD_10MS);

	innopower_vol_init(power->pdev, PWR_NOTIFY_FOR_POWEROFF);

	innopwr_dbg("run finished\n");

	return ret;
}

static int innopwr_restore(struct device *dev)
{
	struct innopower *power = platform_get_drvdata(to_platform_device(dev));

	innopwr_drv_mcufw_sysstat_notify(power, PWR_NOTIFY_FOR_S4_EXIT, VOL_CMD_HOLD_10MS);

	innopwr_exit_sleep(dev, "s4 restore", PWR_NOTIFY_FOR_S4_EXIT);

	return 0;
}

const static struct dev_pm_ops innopwr_pm_ops = {
	.suspend = innopwr_suspend,
	.resume = innopwr_resume,
	.freeze = innopwr_freeze,
	.thaw = innopwr_thaw,
	.poweroff = innopwr_poweroff,
	.restore = innopwr_restore,
};


struct platform_driver innopower_pmbus_driver = {
	.probe = innopower_pmbus_probe,
	.remove = innopower_pmbus_remove,
	.shutdown = innopower_pmbus_shutdown,
	.driver = {
		.name = INNO_POWER_DEVICE_NAME,
		.pm = &innopwr_pm_ops,
	},
	.id_table = innopower_pmbus_device_id_table,
};

#ifdef CONFIG_DRM_INNO_POWER
int innopower_driver_register(void)
#else
static int __init innopower_init(void)
#endif
{
	platform_driver_register(&innopower_pmbus_driver);
	if (i2c_drv) {
		i2c_drv->id_table = inno_powerchip_id_tbl;
		i2c_add_driver(i2c_drv);
	}

#ifdef CONFIG_PM_DEVFREQ
	if (inno_devfreq_gov_register())
		innopwr_err("failed register governor\n");
	inno_input_init();
#endif

	return 0;
}

#ifdef CONFIG_DRM_INNO_POWER
void innopower_driver_unregister(void)
#else
static void __exit innopower_exit(void)
#endif
{
#ifdef CONFIG_PM_DEVFREQ
	inno_input_exit();
	if (inno_devfreq_gov_unregister())
		innopwr_err("failed remove governor\n");
#endif
	if (i2c_drv) {
		i2c_del_driver(i2c_drv);
	}

	platform_driver_unregister(&innopower_pmbus_driver);
}

#ifndef CONFIG_DRM_INNO_POWER
module_init(innopower_init);
module_exit(innopower_exit);
MODULE_LICENSE("Dual MIT/GPL");
#endif
